@@ -30,26 +30,24 @@ import time
 import traceback
 import sys
 import threading
-from typing import Dict, Optional, Tuple, Iterable, Callable, Union, Sequence, Mapping, TYPE_CHECKING
+from typing import Dict, Optional, Tuple, Callable, Union, Sequence, Mapping, TYPE_CHECKING
 from base64 import b64decode, b64encode
-from collections import defaultdict
 import json
 import socket
-from enum import IntEnum
 
 import aiohttp
 from aiohttp import web, client_exceptions
-from aiorpcx import timeout_after, TaskTimeout, ignore_after
+from aiorpcx import ignore_after
 
 from . import util
 from .network import Network
-from .util import (json_decode, to_bytes, to_string, profiler, standardize_path, constant_time_compare, InvalidPassword)
-from .invoices import PR_PAID, PR_EXPIRED
-from .util import log_exceptions, ignore_exceptions, randrange, OldTaskGroup, UserFacingException, JsonRPCError
-from .util import EventListener, event_listener
+from .util import (
+    json_decode, to_bytes, to_string, profiler, standardize_path, constant_time_compare, InvalidPassword,
+    log_exceptions, randrange, OldTaskGroup, UserFacingException, JsonRPCError
+)
 from .wallet import Wallet, Abstract_Wallet
 from .storage import WalletStorage
-from .wallet_db import WalletDB, WalletRequiresSplit, WalletRequiresUpgrade, WalletUnfinished
+from .wallet_db import WalletDB, WalletUnfinished
 from .commands import known_commands, Commands
 from .simple_config import SimpleConfig
 from .exchange_rate import FxThread
@@ -67,9 +65,11 @@ _logger = get_logger(__name__)
 class DaemonNotRunning(Exception):
     pass
 
+
 def get_rpcsock_defaultpath(config: SimpleConfig):
     return os.path.join(config.path, 'daemon_rpc_socket')
 
+
 def get_rpcsock_default_type(config: SimpleConfig):
     if config.RPC_PORT:
         return 'tcp'
@@ -80,9 +80,11 @@ def get_rpcsock_default_type(config: SimpleConfig):
         return 'unix'
     return 'tcp'
 
+
 def get_lockfile(config: SimpleConfig):
     return os.path.join(config.path, 'daemon')
 
+
 def remove_lockfile(lockfile):
     os.unlink(lockfile)
 
@@ -107,15 +109,15 @@ def get_file_descriptor(config: SimpleConfig):
             remove_lockfile(lockfile)
 
 
-
 def request(config: SimpleConfig, endpoint, args=(), timeout: Union[float, int] = 60):
     lockfile = get_lockfile(config)
-    while True:
-        create_time = None
+    for attempt in range(5):
+        create_time = None  # type: Optional[float | int]
         path = None
         try:
             with open(lockfile) as f:
                 socktype, address, create_time = ast.literal_eval(f.read())
+                int(create_time)  # raise if not numeric
                 if socktype == 'unix':
                     path = address
                     (host, port) = "127.0.0.1", 0
@@ -130,6 +132,7 @@ def request(config: SimpleConfig, endpoint, args=(), timeout: Union[float, int]
         server_url = 'http://%s:%d' % (host, port)
         auth = aiohttp.BasicAuth(login=rpc_user, password=rpc_password)
         loop = util.get_asyncio_loop()
+
         async def request_coroutine(
             *, socktype=socktype, path=path, auth=auth, server_url=server_url, endpoint=endpoint,
         ):
@@ -142,15 +145,23 @@ def request(config: SimpleConfig, endpoint, args=(), timeout: Union[float, int]
             async with aiohttp.ClientSession(auth=auth, connector=connector) as session:
                 c = util.JsonRPCClient(session, server_url)
                 return await c.request(endpoint, *args)
+
         try:
             fut = asyncio.run_coroutine_threadsafe(request_coroutine(), loop)
             return fut.result(timeout=timeout)
         except aiohttp.client_exceptions.ClientConnectorError as e:
             _logger.info(f"failed to connect to JSON-RPC server {e}")
-            if not create_time or create_time < time.time() - 1.0:
+            # We cannot communicate with the daemon.
+            # If daemon's creation time is very recent, it might still be starting up.
+            # In any other case, we raise: - too old create_time means daemon is likely dead,
+            #                              - create_time in future means our clock cannot be trusted.
+            if not (create_time <= time.time() <= create_time + 1.0):
                 raise DaemonNotRunning()
-        # Sleep a bit and try again; it might have just been started
+        # Sleep a bit and try again; daemon might have just been started
         time.sleep(1.0)
+    # how did we even get here?! the clock must be going haywire.
+    _logger.error(f"Failed to connect to JSON-RPC server. Exhausted all attempts.")
+    raise DaemonNotRunning()
 
 
 def wait_until_daemon_becomes_ready(*, config: SimpleConfig, timeout=5) -> bool:
@@ -185,12 +196,15 @@ def get_rpc_credentials(config: SimpleConfig) -> Tuple[str, str]:
 class AuthenticationError(Exception):
     pass
 
+
 class AuthenticationInvalidOrMissing(AuthenticationError):
     pass
 
+
 class AuthenticationCredentialsInvalid(AuthenticationError):
     pass
 
+
 class AuthenticatedServer(Logger):
 
     def __init__(self, rpc_user, rpc_password):
@@ -215,7 +229,7 @@ class AuthenticatedServer(Logger):
         if basic != 'Basic':
             raise AuthenticationInvalidOrMissing('UnsupportedType')
         encoded = to_bytes(encoded, 'utf8')
-        credentials = to_string(b64decode(encoded), 'utf8')
+        credentials = to_string(b64decode(encoded, validate=True), 'utf8')
         username, _, password = credentials.partition(':')
         if not (constant_time_compare(username, self.rpc_user)
                 and constant_time_compare(password, self.rpc_password)):
@@ -333,7 +347,10 @@ class CommandsServer(AuthenticatedServer):
         #       "config_options" should have priority.
         if self.daemon.gui_object:
             if hasattr(self.daemon.gui_object, 'new_window'):
-                path = config_options.get('wallet_path') or self.config.get_wallet_path(use_gui_last_wallet=True)
+                if config_options.get(SimpleConfig.NETWORK_OFFLINE.key()) and not self.config.NETWORK_OFFLINE:
+                    raise UserFacingException(
+                        "error: current GUI is running online, so it cannot open a new wallet offline.")
+                path = config_options.get('wallet_path') or self.config.get_wallet_path()
                 self.daemon.gui_object.new_window(path, config_options.get('url'))
                 return True
             else:
@@ -343,7 +360,9 @@ class CommandsServer(AuthenticatedServer):
 
     async def run_cmdline(self, config_options):
         cmdname = config_options['cmd']
-        cmd = known_commands[cmdname]
+        cmd = known_commands.get(cmdname)
+        if not cmd:
+            return f"unknown command: {cmdname}"
         # arguments passed to function
         args = [config_options.get(x) for x in cmd.params]
         # decode json arguments
@@ -352,52 +371,21 @@ class CommandsServer(AuthenticatedServer):
         kwargs = {}
         for x in cmd.options:
             kwargs[x] = config_options.get(x)
-        if 'wallet_path' in cmd.options:
-            kwargs['wallet_path'] = config_options.get('wallet_path')
-        elif 'wallet' in cmd.options:
-            kwargs['wallet'] = config_options.get('wallet_path')
+        if 'wallet_path' in cmd.options or 'wallet' in cmd.options:
+            wallet_path = config_options.get('wallet_path')
+            if len(self.daemon._wallets) > 1 and wallet_path is None:
+                raise UserFacingException("error: wallet not specified")
+            kwargs['wallet_path'] = wallet_path
         func = getattr(self.cmd_runner, cmd.name)
         # execute requested command now.  note: cmd can raise, the caller (self.handle) will wrap it.
         result = await func(*args, **kwargs)
         return result
 
 
-class WatchTowerServer(AuthenticatedServer):
-
-    def __init__(self, network: 'Network', port:int):
-        self.port = port
-        self.config = network.config
-        self.network = network
-        watchtower_user = self.config.WATCHTOWER_SERVER_USER or ""
-        watchtower_password = self.config.WATCHTOWER_SERVER_PASSWORD or ""
-        AuthenticatedServer.__init__(self, watchtower_user, watchtower_password)
-        self.lnwatcher = network.local_watchtower
-        self.app = web.Application()
-        self.app.router.add_post("/", self.handle)
-        self.register_method(self.get_ctn)
-        self.register_method(self.add_sweep_tx)
-
-    async def run(self):
-        self.runner = web.AppRunner(self.app)
-        await self.runner.setup()
-        site = web.TCPSite(self.runner, host='localhost', port=self.port)
-        await site.start()
-        self.logger.info(f"running and listening on port {self.port}")
-
-    async def get_ctn(self, *args):
-        return await self.lnwatcher.get_ctn(*args)
-
-    async def add_sweep_tx(self, *args):
-        return await self.lnwatcher.sweepstore.add_sweep_tx(*args)
-
-
-
-
 class Daemon(Logger):
 
     network: Optional[Network] = None
     gui_object: Optional['gui.BaseElectrumGui'] = None
-    watchtower: Optional['WatchTowerServer'] = None
 
     @profiler
     def __init__(
@@ -415,9 +403,6 @@ class Daemon(Logger):
             fd = get_file_descriptor(config)
             if fd is None:
                 raise Exception('failed to lock daemon; already running?')
-        if 'wallet_path' in config.cmdline_options:
-            self.logger.warning("Ignoring parameter 'wallet_path' for daemon. "
-                                "Use the load_wallet command instead.")
         self._plugins = None  # type: Optional[Plugins]
         self.asyncio_loop = util.get_asyncio_loop()
         if not self.config.NETWORK_OFFLINE:
@@ -460,11 +445,6 @@ class Daemon(Logger):
         self.logger.info(f"starting network.")
         assert not self.config.NETWORK_OFFLINE
         assert self.network
-        # server-side watchtower
-        if watchtower_port := self.config.WATCHTOWER_SERVER_PORT:
-            self.watchtower = WatchTowerServer(self.network, watchtower_port)
-            asyncio.run_coroutine_threadsafe(self.taskgroup.spawn(self.watchtower.run), self.asyncio_loop)
-
         self.network.start(jobs=[self.fx.run])
         # prepare lightning functionality, also load channel db early
         if self.config.LIGHTNING_USE_GOSSIP:
@@ -473,13 +453,21 @@ class Daemon(Logger):
     @staticmethod
     def _wallet_key_from_path(path) -> str:
         """This does stricter path standardization than 'standardize_path'.
-        It is used for keying the _wallets dict, but not for the actual filesystem operations. (see #8495)
+        It is used for keying the _wallets dict,
+        but MUST NOT be used as a *path* for the actual filesystem operations. (see #8495)
         """
         path = standardize_path(path)
-        # also resolve symlinks and windows network mounts/etc:
+        # The extra normalisation makes it even harder to open the same wallet file multiple times simultaneously.
+        # - "realpath" resolves symlinks:
+        #   note: the path returned by realpath has been observed NOT to work for FS operations!
+        #         (e.g. for Cryptomator WinFSP/FUSE mounts, see #8495).
+        #         It is okay for us to use it for computing a canonical wallet *key*, but cannot be used as a path!
         path = os.path.realpath(path)
+        # - "normcase" does Windows-specific case and slash normalisation:
         path = os.path.normcase(path)
-        return str(path)
+        # - prepend header to break usage of wallet keys as fs paths
+        header = "WALLETKEY-"
+        return header + str(path)
 
     def with_wallet_lock(func):
         def func_wrapper(self: 'Daemon', *args, **kwargs):
@@ -489,16 +477,26 @@ class Daemon(Logger):
 
     @with_wallet_lock
     def load_wallet(self, path, password, *, upgrade=False) -> Optional[Abstract_Wallet]:
+        assert password != ''
         path = standardize_path(path)
         wallet_key = self._wallet_key_from_path(path)
         # wizard will be launched if we return
         if wallet := self._wallets.get(wallet_key):
             return wallet
         wallet = self._load_wallet(path, password, upgrade=upgrade, config=self.config)
-        wallet.start_network(self.network)
+        if self.network:
+            wallet.start_network(self.network)
+        elif wallet.lnworker:
+            # in offline mode, we need to trigger callbacks
+            coro = wallet.lnworker.lnwatcher.trigger_callbacks(requires_synchronizer=False)
+            asyncio.run_coroutine_threadsafe(coro, self.asyncio_loop)
         self.add_wallet(wallet)
+        if self.config.get('wallet_path') is None:
+            self.config.CURRENT_WALLET = path
+        self.update_recently_opened_wallets(path)
         return wallet
 
+
     @staticmethod
     @profiler
     def _load_wallet(
@@ -509,7 +507,7 @@ class Daemon(Logger):
             config: SimpleConfig,
     ) -> Optional[Abstract_Wallet]:
         path = standardize_path(path)
-        storage = WalletStorage(path)
+        storage = WalletStorage(path, allow_partial_writes=config.WALLET_PARTIAL_WRITES)
         if not storage.file_exists():
             raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)
         if storage.is_encrypted():
@@ -542,6 +540,7 @@ class Daemon(Logger):
         self.stop_wallet(path)
         if os.path.exists(path):
             os.unlink(path)
+            self.update_recently_opened_wallets(path, remove=True)
             return True
         return False
 
@@ -554,14 +553,23 @@ class Daemon(Logger):
     @with_wallet_lock
     async def _stop_wallet(self, path: str) -> bool:
         """Returns True iff a wallet was found."""
+        path = standardize_path(path)
         wallet_key = self._wallet_key_from_path(path)
         wallet = self._wallets.pop(wallet_key, None)
         if not wallet:
             return False
         await wallet.stop()
+        if self.config.get('wallet_path') is None:
+            wallet_paths = [w.db.storage.path for w in self._wallets.values()
+                            if w.db.storage and w.db.storage.path]
+            if self.config.CURRENT_WALLET == path and wallet_paths:
+                self.config.CURRENT_WALLET = wallet_paths[0]
         return True
 
     def run_daemon(self):
+        if 'wallet_path' in self.config.cmdline_options:
+            self.logger.warning("Ignoring parameter 'wallet_path' for daemon. "
+                                "Use the load_wallet command instead.")
         # init plugins
         self._plugins = Plugins(self.config, 'cmdline')
         # block until we are stopping
@@ -704,3 +712,14 @@ class Daemon(Logger):
         self._check_password_for_directory(
             old_password=old_password, new_password=new_password, wallet_dir=wallet_dir)
         return True
+
+    def update_recently_opened_wallets(self, wallet_path, *, remove: bool = False):
+        recent = self.config.RECENTLY_OPEN_WALLET_FILES or []
+        if wallet_path in recent:
+            recent.remove(wallet_path)
+        if not remove:
+            recent.insert(0, wallet_path)
+            recent = [path for path in recent if os.path.exists(path)]
+            recent = recent[:5]
+        self.config.RECENTLY_OPEN_WALLET_FILES = recent
+        util.trigger_callback('recently_opened_wallets_update')
